
class WhodapError(Exception):
    ...


class NotFoundError(WhodapError):
    ...


class RateLimitError(WhodapError):
    ...


class MalformedQueryError(WhodapError):
    ...


class BadStatusCode(WhodapError):
    ...


from abc import ABC, abstractmethod
from datetime import datetime
from json import dumps, loads
from types import SimpleNamespace
from typing import Dict, Any, List, Union

from .utils import WHOISKeys, RDAPVCardKeys

REDACTED = 'REDACTED FOR PRIVACY'


class RDAPResponse(SimpleNamespace, ABC):
    """
    Abstract class representing an RDAP Response
    """

    def __init__(self, *args, **kwargs):
        super().__init__(**kwargs)

    def __str__(self):
        return self.to_json(indent=2)

    def __repr__(self):
        return self.to_json(indent=2)

    @abstractmethod
    def to_json(self, **kwargs):
        ...

    @abstractmethod
    def to_dict(self, **kwargs):
        ...

    @staticmethod
    def _convert_date(ds: str) -> Union[str, datetime]:
        """
        Utility for converting known RDAP date strings
        into Python datetime objects.

        :param ds: a date string
        :return: a datetime object or the original string
        """
        known_rdap_formats = (
            '%Y-%m-%dT%H:%M:%S',
            '%Y-%m-%dT%H:%M:%SZ',
            '%Y-%m-%dT%H:%M:%S.%f%z',
            '%Y-%m-%dT%H:%M:%S%z',
            # https://stackoverflow.com/questions/53291250/python-3-6-datetime-strptime-returns-error-while-python-3-7-works-well
            '%Y-%m-%dT%H:%M:%S.%fZ'
        )
        for date_format in known_rdap_formats:
            try:
                # try every date format used by RDAP
                v = datetime.strptime(ds, date_format)
                return v
            except ValueError:
                continue
        # return original date string if unsuccessful
        return ds

    def _convert_list(self, ls: List[Any]) -> List[Any]:
        """
        Iterates over the given list checking for nested RDAPResponses;
        Recursively calls itself when it encounters another list otherwise
        converts RDAPResponses and appends values.

        :param ls: any list
        :return: another list
        """
        converted = []
        for obj in ls:
            if isinstance(obj, type(self)):
                converted.append(self._convert_self_to_dict(obj))
            elif isinstance(obj, list):
                converted.append(self._convert_list(obj))
            else:
                converted.append(obj)
        return converted

    def _convert_self_to_dict(self, rdr: 'RDAPResponse') -> Dict[str, Any]:
        """
        Converts the RDAPResponse to a dictionary.
        Recursively calls itself to convert nested RDAPResponse.

        :param rdr: an instance of RDAPResponse
        :return: the RDAPResponse converted to a dictionary
        """
        converted = {}
        for key, value in rdr.__dict__.items():
            if isinstance(value, list):
                converted[key] = self._convert_list(value)
            elif isinstance(value, type(self)):
                converted[key] = self._convert_self_to_dict(value)
            else:
                converted[key] = value
        return converted

    @classmethod
    def from_json(cls, json: Union[str, bytes]):
        """
        Initializes an instance of DomainResponse from the
        JSON output of an RDAP HTTP query.

        :param json: JSON Response from the RDAP server
        :return: an RDAPResponse
        """
        return loads(json, object_hook=lambda d: cls(**d))


class DomainResponse(RDAPResponse):

    def __getattribute__(self, item):
        """
        Converts and returns an "eventDate" value to a datetime object;
        otherwise returns the value of the given attribute
        """
        val = super().__getattribute__(item)
        if item == 'eventDate':
            return self._convert_date(val)
        return val

    @staticmethod
    def _encoder(x: Any):
        """
        JSON encoding helper for "datetime" objects.

        :param x: Any
        :return: "datetime" as an iso formatted string or Any
        """
        if isinstance(x, datetime):
            return x.isoformat()
        return x

    def to_json(self, **kwargs) -> str:
        """
        Converts the DomainResponse to a JSON string.

        :param kwargs: arguments to be passed to `json.dumps`
        :return: JSON string
        """
        if not kwargs.get('default'):
            kwargs['default'] = self._encoder
        return dumps(self.to_dict(), **kwargs)

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts the DomainResponse to a dictionary.
        """
        return self._convert_self_to_dict(self)

    def to_whois_json(self, **kwargs) -> str:
        """
        Converts the DomainResponse to a WHOIS JSON string.

        :param kwargs: arguments to be passed to `json.dumps`
        :return: JSON string
        """
        if not kwargs.get('default'):
            kwargs['default'] = self._encoder
        return dumps(self.to_whois_dict(), **kwargs)

    def to_whois_dict(self) -> Dict[WHOISKeys, Union[str, List[str], datetime, None]]:
        """
        Returns the DomainResponse as "flattened" WHOIS dictionary;
        does not modify the original DomainResponse object.

        :return: dict with WHOIS keys
        """
        flat = {}

        # traverse and extract information from RDAP fields
        if getattr(self, 'nameservers', None):
            flat_nameservers = {'nameservers': []}
            for obj in self.nameservers:
                flat_nameservers['nameservers'].append(obj.ldhName)
            flat.update(flat_nameservers)

        if getattr(self, 'status', None):
            flat.update({'status': self.status})

        if getattr(self, 'events', None):
            flat.update(self._flat_dates(self.events))

        if getattr(self, 'entities', None):
            flat.update(self._flat_entities(self.entities))

        # convert dict keys over to "WHOISKeys"
        flat = self._construct_flat_dict(flat)

        # add domain name
        flat[WHOISKeys.DOMAIN_NAME] = self.ldhName

        # add dnssec after ensuring that it exists
        if getattr(self, 'secureDNS', None) and \
                getattr(self.secureDNS, 'delegationSigned', None):
            flat[WHOISKeys.DNSSEC] = self.secureDNS.delegationSigned

        return flat

    def _flatten_list(self, ls: List[Any]):
        """
        Recursively flattens the input list.

        :param ls: any list
        :return: a flattened list
        """
        flattened = []
        for i in ls:
            if isinstance(i, list):
                flattened.extend(self._flatten_list(i))
            else:
                flattened.append(i)
        return flattened

    @staticmethod
    def _flat_dates(events: List[SimpleNamespace]) -> Dict[str, datetime]:
        """
        Returns the list of events as a flattened dict of date keys and values

        :param events: list of "events" from the RDAP response
        :return: dictionary of "date" key value pairs
        """
        dates = dict([(event.eventAction, event.eventDate) for event in events])
        return dates

    def _flat_entities(self, entities: List[SimpleNamespace]) -> Dict[str, Dict[str, str]]:
        entities_dict = {}
        for entity in entities:
            ent_dict = {}
            # check for redacted information
            mark_redacted = False
            if hasattr(entity, 'remarks'):
                for remark in entity.remarks:
                    if hasattr(remark, 'title') and 'redact' in remark.title.lower():
                        mark_redacted = True
            # check for nested entities
            if hasattr(entity, 'entities'):
                # recursive call for nested entities
                ent_dict = self._flat_entities(entity.entities)
                entities_dict.update(ent_dict)
            # iterate through vCard array
            if hasattr(entity, 'vcardArray'):
                for vcard in entity.vcardArray[-1]:
                    # vCard represents information about an individual or entity.
                    vcard_type = vcard[0]
                    vcard_value = vcard[-1]
                    # check for organization
                    if vcard_type == RDAPVCardKeys.ORG:
                        ent_dict['org'] = vcard_value
                    # check for email
                    elif vcard_type == RDAPVCardKeys.EMAIL:
                        ent_dict['email'] = vcard_value
                    # check for name
                    elif vcard_type == RDAPVCardKeys.FN:
                        ent_dict['name'] = vcard_value
                    # check for address
                    elif vcard_type == RDAPVCardKeys.ADR:
                        values = self._flatten_list(vcard_value)
                        address_string = ', '.join([v for v in values if v])
                        ent_dict['address'] = address_string.lstrip()
                    # check for contact
                    elif vcard_type == RDAPVCardKeys.TEL:
                        # check the "type" of "tel" vcard (either voice or fax):
                        # vcard looks like: ['tel', {"type": "voice"}, 'uri', 'tel:0000000']
                        if hasattr(vcard[1], 'type'):
                            contact_type = vcard[1].to_dict().get('type')
                            if contact_type == 'voice':
                                ent_dict['phone'] = vcard_value
                            elif contact_type == 'fax':
                                ent_dict['fax'] = vcard_value
                        else:
                            ent_dict['phone'] = vcard_value
                        
            # add roles for this entity
            for role in entity.roles:
                if mark_redacted:
                    for key in ('address', 'phone', 'name', 'org', 'email', 'fax'):
                        if not ent_dict.get(key):
                            ent_dict[key] = REDACTED
                # save the information under this "role"
                entities_dict[role.lower()] = ent_dict

        # return parsed entities dict
        return entities_dict

    @staticmethod
    def _construct_flat_dict(parsed: Dict[str, Any]) -> Dict[WHOISKeys, Any]:
        converted = {
            WHOISKeys.ABUSE_EMAIL: parsed.get('abuse', {}).get('email'),
            WHOISKeys.ABUSE_PHONE: parsed.get('abuse', {}).get('phone'),
            WHOISKeys.ADMIN_NAME: parsed.get('administrative', {}).get('name'),
            WHOISKeys.ADMIN_ORG: parsed.get('administrative', {}).get('org'),
            WHOISKeys.ADMIN_EMAIL: parsed.get('administrative', {}).get('email'),
            WHOISKeys.ADMIN_ADDRESS: parsed.get('administrative', {}).get('address'),
            WHOISKeys.ADMIN_PHONE: parsed.get('administrative', {}).get('phone'),
            WHOISKeys.ADMIN_FAX: parsed.get('administrative', {}).get('fax'),
            WHOISKeys.BILLING_NAME: parsed.get('billing', {}).get('name'),
            WHOISKeys.BILLING_ORG: parsed.get('billing', {}).get('org'),
            WHOISKeys.BILLING_EMAIL: parsed.get('billing', {}).get('email'),
            WHOISKeys.BILLING_ADDRESS: parsed.get('billing', {}).get('address'),
            WHOISKeys.BILLING_PHONE: parsed.get('billing', {}).get('phone'),
            WHOISKeys.BILLING_FAX: parsed.get('billing', {}).get('fax'),
            WHOISKeys.REGISTRANT_NAME: parsed.get('registrant', {}).get('name'),
            WHOISKeys.REGISTRANT_ORG: parsed.get('registrant', {}).get('organization'),
            WHOISKeys.REGISTRANT_EMAIL: parsed.get('registrant', {}).get('email'),
            WHOISKeys.REGISTRANT_ADDRESS: parsed.get('registrant', {}).get('address'),
            WHOISKeys.REGISTRANT_PHONE: parsed.get('registrant', {}).get('phone'),
            WHOISKeys.REGISTRANT_FAX: parsed.get('registrant', {}).get('fax'),
            WHOISKeys.REGISTRAR_NAME: parsed.get('registrar', {}).get('name'),
            WHOISKeys.REGISTRAR_EMAIL: parsed.get('registrar', {}).get('email'),
            WHOISKeys.REGISTRAR_ADDRESS: parsed.get('registrar', {}).get('address'),
            WHOISKeys.REGISTRAR_PHONE: parsed.get('registrar', {}).get('phone'),
            WHOISKeys.REGISTRAR_FAX: parsed.get('registrar', {}).get('fax'),
            WHOISKeys.TECHNICAL_NAME: parsed.get('technical', {}).get('name'),
            WHOISKeys.TECHNICAL_ORG: parsed.get('technical', {}).get('org'),
            WHOISKeys.TECHNICAL_EMAIL: parsed.get('technical', {}).get('email'),
            WHOISKeys.TECHNICAL_ADDRESS: parsed.get('technical', {}).get('address'),
            WHOISKeys.TECHNICAL_PHONE: parsed.get('technical', {}).get('phone'),
            WHOISKeys.TECHNICAL_FAX: parsed.get('technical', {}).get('fax'),
            WHOISKeys.CREATED_DATE: parsed.get('registration'),
            WHOISKeys.UPDATED_DATE: parsed.get('last update') or parsed.get('last changed'),
            WHOISKeys.EXPIRES_DATE: parsed.get('expiration'),
            WHOISKeys.STATUS: parsed.get('status'),
            WHOISKeys.NAMESERVERS: parsed.get('nameservers')
        }
        return converted


class IPv4Response(RDAPResponse):

    def to_json(self, **kwargs):
        ...

    def to_dict(self, **kwargs):
        ...

    def get_geo_location(self):
        ...


class IPv6Response(RDAPResponse):

    def to_json(self, **kwargs):
        ...

    def to_dict(self, **kwargs):
        ...


class ASNResponse(RDAPResponse):

    def to_json(self, **kwargs):
        ...

    def to_dict(self, **kwargs):
        ...


from enum import Enum


class RDAPStatusCodes(int, Enum):
    """
    Status Code meanings for the RDAP protocol
    """
    POSITIVE_ANSWER_200 = 200
    NEGATIVE_ANSWER_404 = 404
    MALFORMED_QUERY_400 = 400
    RATE_LIMIT_429 = 429
    REDIRECT_302 = 302



import sys
import posixpath
import ipaddress
from abc import ABC, abstractmethod
from typing import Dict, Any, Union, Optional
from contextlib import contextmanager

# different installs for async contextmanager based on python version
if sys.version_info.major == 3 and sys.version_info.minor < 7:
    from async_generator import asynccontextmanager
else:
    from contextlib import asynccontextmanager

import httpx

from .codes import RDAPStatusCodes
from .errors import RateLimitError, NotFoundError, MalformedQueryError, BadStatusCode
from .response import DomainResponse


class RDAPClient(ABC):
    _iana_publication_key: str = 'publication'
    _iana_verison_key: str = 'version'
    _iana_services_key: str = 'services'
    _iana_uri: str = None

    def __init__(self, httpx_client: Union[httpx.Client, httpx.AsyncClient]):
        self.httpx_client = httpx_client
        self.version: str = ''
        self.publication: str = ''

    @classmethod
    @abstractmethod
    def new_client_context(cls, httpx_client):
        ...

    @classmethod
    @abstractmethod
    def new_client(cls, httpx_client):
        ...

    @classmethod
    @abstractmethod
    async def new_aio_client_context(cls, httpx_client):
        ...

    @classmethod
    @abstractmethod
    async def new_aio_client(cls, httpx_client):
        ...

    @abstractmethod
    def lookup(self):
        ...

    @abstractmethod
    async def aio_lookup(self):
        ...

    @staticmethod
    @abstractmethod
    def _build_query_uri() -> str:
        ...

    def _get_request(self, uri: str) -> httpx.Response:
        return self.httpx_client.get(uri)

    async def _aio_get_request(self, uri: str) -> httpx.Response:
        return await self.httpx_client.get(uri)

    @staticmethod
    def _check_status_code(status_code: int) -> None:
        if status_code == RDAPStatusCodes.POSITIVE_ANSWER_200:
            return None
        elif status_code == RDAPStatusCodes.MALFORMED_QUERY_400:
            raise MalformedQueryError(
                f"Malformed query: {RDAPStatusCodes.MALFORMED_QUERY_400}")
        elif status_code == RDAPStatusCodes.NEGATIVE_ANSWER_404:
            raise NotFoundError(
                f"Domain not found: {RDAPStatusCodes.NEGATIVE_ANSWER_404}")
        elif status_code == RDAPStatusCodes.RATE_LIMIT_429:
            raise RateLimitError(
                f"Too many requests: {RDAPStatusCodes.RATE_LIMIT_429}")
        else:
            raise BadStatusCode(f"Status code <{status_code}>")


class DNSClient(RDAPClient):
    # IANA DNS
    _iana_uri: str = 'https://data.iana.org/rdap/dns.json'

    def __init__(self, httpx_client: Union[httpx.Client, httpx.AsyncClient]):
        super(DNSClient, self).__init__(httpx_client)
        self.iana_dns_server_map: Dict[str, str] = {}

    @classmethod
    @contextmanager
    def new_client_context(cls, httpx_client: Optional[httpx.Client] = None):
        """
        Contextmanager for instantiating a Synchronous DNSClient

        :httpx_client: pre-configured instance of `httpx.Client`
        :return: yields the initialized DNSClient
        """
        dns_client = cls(httpx_client or httpx.Client())
        try:
            iana_dns_info = dns_client.get_iana_dns_info()
            dns_client.set_iana_dns_info(iana_dns_info)
            yield dns_client
        except:
            raise
        finally:
            if not dns_client.httpx_client.is_closed:
                dns_client.httpx_client.close()

    @classmethod
    def new_client(cls, httpx_client: Optional[httpx.Client] = None):
        """
        Classmethod for instantiating an synchronous instance of DNSClient

        :httpx_client: pre-configured instance of `httpx.Client`
        :return: DNSClient with a sync httpx_client
        """
        # init the client with a default httpx.Client if one is not provided
        dns_client = cls(httpx_client or httpx.Client())
        # load the dns server information from IANA
        iana_dns_info = dns_client.get_iana_dns_info()
        # parse and save the server information
        dns_client.set_iana_dns_info(iana_dns_info)
        # return the loaded client
        return dns_client

    @classmethod
    @asynccontextmanager
    async def new_aio_client_context(cls, httpx_client: Optional[httpx.AsyncClient] = None):
        """
        Contextmanager for instantiating an Asynchronous DNSClient

        :httpx_client: Optional pre-configured instance of `httpx.AsyncClient`
        :return: yields the initialized DNSClient
        """
        dns_client = cls(httpx_client or httpx.AsyncClient())
        try:
            iana_dns_info = await dns_client.aio_get_iana_dns_info()
            dns_client.set_iana_dns_info(iana_dns_info)
            yield dns_client
        except:
            raise
        finally:
            if not dns_client.httpx_client.is_closed:
                await dns_client.httpx_client.aclose()

    @classmethod
    async def new_aio_client(cls, httpx_client: Optional[httpx.AsyncClient] = None):
        """
        Classmethod for instantiating an asynchronous instance of DNSClient

        :httpx_client: pre-configured instance of `httpx.AsyncClient`
        :return: DNSClient with an async httpx_client
        """
        dns_client = cls(httpx_client or httpx.AsyncClient())
        iana_dns_info = await dns_client.aio_get_iana_dns_info()
        dns_client.set_iana_dns_info(iana_dns_info)
        return dns_client

    def get_iana_dns_info(self):
        response = self._get_request(self._iana_uri)
        return response.json()

    async def aio_get_iana_dns_info(self):
        response = await self._aio_get_request(self._iana_uri)
        return response.json()

    @staticmethod
    def _build_query_uri(rdap_uri: str, domain: str) -> str:
        return posixpath.join(rdap_uri, 'domain', domain.lstrip('/'))

    def lookup(self, domain: str, tld: str, auth_ref: str = None) -> DomainResponse:
        """
        Performs an RDAP domain lookup.

        First, finds the appropriate server for the top level domain,
        sends an HTTP request to the server, parses the response for a more authoritative source,
        sends an additional HTTP request to the more authoritative source, and finally
        encapsulates the HTTP response into a DomainResponse object.

        :param domain: the domain name
        :param tld: the top level domain
        :param auth_ref: optional authoritative url for the given TLD
        :return: instance of DomainResponse
        """
        domain_and_tld = domain + '.' + tld
        # if an authoritative url is provided; use it
        if auth_ref:
            query_url = self._build_query_uri(auth_ref, domain_and_tld)
            resp = self._get_request(query_url)
            self._check_status_code(resp.status_code)
            return DomainResponse.from_json(resp.text)
        # start with looking up server in the IANA list
        server_url = self.iana_dns_server_map.get(tld)
        if not server_url:
            raise NotImplementedError(f'No RDAP Server for ".{tld.upper()}"')
        # hit the server found in the IANA list
        query_url = self._build_query_uri(server_url, domain_and_tld)
        response = self._get_request(query_url)
        self._check_status_code(response.status_code)
        domain_response = DomainResponse.from_json(response.text)
        # try to extract an authoritative server for this domain
        if hasattr(domain_response, 'links'):
            authoritative_url = domain_response.links[-1].href
            # avoid redundant connections
            if authoritative_url.lower() != query_url.lower():
                resp = self._get_request(authoritative_url)
                self._check_status_code(resp.status_code)
                return DomainResponse.from_json(resp.text)
            else:
                return domain_response
        else:
            return domain_response

    async def aio_lookup(self, domain: str, tld: str, auth_ref: str = None) -> DomainResponse:
        """
        Performs an RDAP domain lookup.

        First, finds the appropriate server for the top level domain,
        sends an HTTP request to the server, parses the response for a more authoritative source,
        sends an additional HTTP request to the more authoritative source, and finally
        encapsulates the HTTP response into a DomainResponse object.

        :param domain: the domain name
        :param tld: the top level domain
        :param auth_ref: optional authoritative url for the given TLD
        :return: instance of DomainResponse
        """
        domain_and_tld = domain + '.' + tld
        if auth_ref:
            query_url = self._build_query_uri(auth_ref, domain_and_tld)
            resp = await self._aio_get_request(query_url)
            self._check_status_code(resp.status_code)
            return DomainResponse.from_json(resp.read())
        server_url = self.iana_dns_server_map.get(tld)
        if not server_url:
            raise NotImplementedError(f'Could not find RDAP server for .{tld.upper()} domains')
        query_url = self._build_query_uri(server_url, domain_and_tld)
        response = await self._aio_get_request(query_url)
        self._check_status_code(response.status_code)
        domain_response = DomainResponse.from_json(response.read())
        if hasattr(domain_response, 'links'):
            authoritative_url = domain_response.links[-1].href
            if authoritative_url.lower() != query_url.lower():
                resp = await self._aio_get_request(authoritative_url)
                self._check_status_code(resp.status_code)
                return DomainResponse.from_json(resp.read())
            else:
                return domain_response
        else:
            return domain_response

    def set_iana_dns_info(self, iana_dns_map: Dict[str, Any]) -> None:
        """
        Populates the DNSClient's `iana_dns_server_map` attribute with
        the server information found in the given `iana_dns_map`.

        :param iana_dns_map: Server information retrieved from `self._iana_url`
        :return: None
        """
        self.publication = iana_dns_map.get(self._iana_publication_key)
        self.version = iana_dns_map.get(self._iana_verison_key)
        tld_server_map = {}
        for tlds, server in iana_dns_map.get(self._iana_services_key):
            for tld in tlds:
                tld_server_map[tld] = server[0]
        self.iana_dns_server_map = tld_server_map


class IPv4Client(RDAPClient):
    # IANA IPv4
    _iana_uri: str = 'https://data.iana.org/rdap/ipv4.json'
    _arin_registry_uri: str = 'https://rdap.arin.net/registry/ip/'

    def __init__(self):
        super(IPv4Client, self).__init__()
        self.iana_ipv4_server_map: Dict[str, str] = {}

    @classmethod
    def new_client(cls):
        ...

    @classmethod
    async def new_aio_client(cls):
        ...

    def lookup(self):
        ...

    async def aio_lookup(self):
        ...

    @staticmethod
    def _build_query_uri(rdap_uri: str, ip_address: str) -> str:
        return posixpath.join(rdap_uri, ip_address)

    def _set_ipv4_server_map(self, iana_ipv4_map: Dict[str, Any]):
        ...


class IPv6Client(RDAPClient):
    # IANA IPv6
    ...

    @classmethod
    def new_client(cls):
        ...

    @classmethod
    async def new_aio_client(cls):
        ...

    def lookup(self):
        ...

    async def aio_lookup(self):
        ...

    @staticmethod
    def _build_query_uri() -> str:
        ...


class ASNClient(RDAPClient):
    # IANA ASN
    ...

    @classmethod
    def new_client(cls):
        ...

    @classmethod
    async def new_aio_client(cls):
        ...

    def lookup(self):
        ...

    async def _aio_load_from_iana(self):
        ...

    @staticmethod
    def _build_query_uri() -> str:
        ...



from enum import Enum


class RDAPVCardKeys(str, Enum):
    ADR = 'adr'
    EMAIL = 'email'
    FN = 'fn'
    ORG = 'org'
    TEL = 'tel'

    @staticmethod
    def values():
        return [rdk.value for rdk in RDAPVCardKeys]


class WHOISKeys(str, Enum):
    """
    WHOIS field names
    """
    DOMAIN_NAME = 'domain_name'
    NAMESERVERS = 'nameservers'
    DNSSEC = 'dnssec'
    STATUS = 'status'
    CREATED_DATE = 'created_date'
    EXPIRES_DATE = 'expires_date'
    UPDATED_DATE = 'updated_date'
    ABUSE_EMAIL = 'abuse_email'
    ABUSE_PHONE = 'abuse_phone'
    ADMIN_NAME = 'admin_name'
    ADMIN_ORG = 'admin_organization'
    ADMIN_EMAIL = 'admin_email'
    ADMIN_ADDRESS = 'admin_address'
    ADMIN_PHONE = 'admin_phone'
    ADMIN_FAX = 'admin_fax'
    BILLING_NAME = 'billing_name'
    BILLING_ORG = 'billing_organization'
    BILLING_EMAIL = 'billing_email'
    BILLING_ADDRESS = 'billing_address'
    BILLING_PHONE = 'billing_phone'
    BILLING_FAX = 'billing_fax'
    REGISTRAR_NAME = 'registrar_name'
    REGISTRAR_EMAIL = 'registrar_email'
    REGISTRAR_ADDRESS = 'registrar_address'
    REGISTRAR_PHONE = 'registrar_phone'
    REGISTRAR_FAX = 'registrar_fax'
    REGISTRANT_NAME = 'registrant_name'
    REGISTRANT_ORG = 'registrant_organization'
    REGISTRANT_EMAIL = 'registrant_email'
    REGISTRANT_ADDRESS = 'registrant_address'
    REGISTRANT_PHONE = 'registrant_phone'
    REGISTRANT_FAX = 'registrant_fax'
    TECHNICAL_NAME = 'technical_name'
    TECHNICAL_ORG = 'technical_organization'
    TECHNICAL_EMAIL = 'technical_email'
    TECHNICAL_ADDRESS = 'technical_address'
    TECHNICAL_PHONE = 'technical_phone'
    TECHNICAL_FAX = 'technical_fax'

    def __str__(self):
        return self.value

    def __repr__(self):
        return self.value



from setuptools import setup
import pathlib


base_dir = pathlib.Path(__file__).parent.resolve()
long_description = (base_dir / 'README.md').read_text(encoding='utf-8')


def get_version(location: str) -> str:
    with open((base_dir / location).absolute().resolve()) as file:
        for line in file.readlines():
            if line.startswith('__version__'):
                return line.split(' = ')[-1].strip().replace("'", "")
        else:
            raise RuntimeError('Unable to find version string.')


setup(
    name='whodap',
    version=get_version('whodap/__init__.py'),
    description='Simple RDAP Utility for Python',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/pogzyb/whodap',
    author='Joseph Obarzanek',
    author_email='pogzyb@umich.edu',
    python_requires='>=3.6, <4',
    keywords='security, whois, rdap, research',
    install_requires=[
        'httpx>=0.18.1',
        'async_generator>=1.10; python_version < "3.7.0"'
    ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development',
        'Topic :: Security',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3 :: Only',
    ],
    packages=['whodap'],
    package_dir={'whodap': 'whodap'},
)



from typing import Optional

from httpx import Client, AsyncClient

from .client import DNSClient
from .response import DomainResponse

__all__ = ['aio_lookup_domain', 'lookup_domain', 'DNSClient']
__version__ = '0.1.1'


def lookup_domain(domain: str,
                  tld: str,
                  httpx_client: Optional[Client] = None) -> DomainResponse:
    """
    Convenience function that instantiates a DNSClient,
    submits an RDAP query for the given domain, and returns
    the result as a DomainResponse.

    :param domain: the domain name to lookup
    :param tld: the top level domain (e.g. "com", "net", "buzz")
    :param httpx_client: Custom, pre-configured instance `httpx.Client`
    :return: an instance of DomainResponse
    """
    dns_client = DNSClient.new_client(httpx_client)
    response = dns_client.lookup(domain, tld)
    if not httpx_client and not dns_client.httpx_client.is_closed:
        dns_client.httpx_client.close()
    return response


async def aio_lookup_domain(domain: str,
                            tld: str,
                            httpx_client: Optional[AsyncClient] = None) -> DomainResponse:
    """
    Async-compatible convenience function that instantiates
    a DNSClient, submits an RDAP query for the given domain,
    and returns the result as a DomainResponse.

    :param domain: the domain name to lookup
    :param tld: the top level domain (e.g. "com", "net", "buzz")
    :param httpx_client: Custom, pre-configured instance `httpx.AsyncClient`
    :return: an instance of DomainResponse
    """
    dns_client = await DNSClient.new_aio_client(httpx_client)
    response = await dns_client.aio_lookup(domain, tld)
    if not httpx_client and not dns_client.httpx_client.is_closed:
        await dns_client.httpx_client.close()
    return response
